from threeD import *
import math
import copy
#■光線(Ray)クラス
class Ray:
    def __init__(self, o, d):
        self.o=o; self.d=d
#■球(Ball)クラス
class Ball(Point):
    def __init__(self, o=Point(0,0,0), r=0):
        #if o==0: o=Point(0,0,0)
        self.o=o
        self.r=r
    def hit(self, VR):#交点があるかどうか
        ds=(VR.o-self.o)*VR.d
        d2=ds*ds-(VR.o-self.o)*(VR.o-self.o)+self.r*self.r;
        if d2<0 : return [1E+8]
        t= float(-ds - math.sqrt(d2))
        P=VR.o+ VR.d*t; N=unit(P-self.o)
        return [t, copy.copy(P), copy.copy(N)]
    def hitV(self, VR):
        ds=VR.d*(VR.o-o)
        d2=ds*ds-(VR.o-self.o)*(VR.o-self.o)+self.r*self.r
        if d2<0 :  return 1E+8
        else :return float( -ds - math.sqrt(d2))
#■壁(Wall)クラス(床もこのクラスとみなす)
class Wall(Polygon):
    def __init__(self, pi):
        self.polygon=Polygon(pi)
        self.nv=unit((self.polygon.p[0]-self.polygon.p[2])%
                     (self.polygon.p[1]-self.polygon.p[3]))
    def hit(self, VR):
        t=self.nv*(self.polygon.p[1]-VR.o)/(self.nv*VR.d) #距離
        if t<=0 : return [1E+8]
        P=VR.o+VR.d*t; N=self.nv
        for i in range(len(self.polygon.p)-1):
            if(N*((self.polygon.p[i+1]-self.polygon.p[i])%
                  (P-self.polygon.p[i])) < 0):
                return [1E+8]
        return [t, copy.copy(P), copy.copy(N)]
#■各種属性(Attribute)クラス
class Attribute(Point):
    def __init__(self, C=Color(255,255,255), D=0.7,S=0.8,E=0.2,
                 R=0,N=0,T=0.5):
        if isinstance(C, Color):
            self.c=C                            # 色 
            self.tx=[Color(0,0,0)]              # テクスチャ
            self.kd=D; self.ks=S; self.ke=E     # 拡散反射率, 鏡面反射率, 環境光係数
            self.kr=R; self.kn=N; self.kt=T     # 反射率, 屈折率, 透過率
            self.wi=0; self.hi=0                # ウィンドウ・サイズ
            self.rn=[1]
        elif isinstance(C,Attribute):
            self.c=a.c
            self.tx=a.tx
            self.kd=a.kd;self.ks=a.ks; self.ke=a.ke
            self.kr=a.kr; self.kn=a.kn; self.kt=a.kt
            self.wi=a.wi; self.hi=a.hi
            self.rn=a.rn
    def shading(self, V, L, N, A, s):
        RV=L-N*2*(L*N)
        D1=(-N*L)/(length(N)*lengh(L)) #拡散および環境光
        if D1<0 : D1=0
        D1=A.kd * self.s * D1 + A.ke
        D2=self.s*A.ks*((-RV*V))/(len(RV)*len(V))    # 反射
        if D2<0 : D2=0
        R= int(D1*(float(A.c.R))+D2*255)
        if R>255 : R=255
        G= int(D1*(float(A.c.G))+D2*255)
        if G>255 : G=255
        B= int(D1*(float(A.c.B))+D2*255)
        if B>255 : B=255
        return Color(R,G,B)

#■図形のバラエティをオブジェクトとして格納するためのクラス
class Object:
    WALL    =0x10
    BALL    =0x20
    FREE    =0x40
    def __init__(self, X,a):
        if isinstance(X,Wall):
            self.w=X; self.Type=self.WALL; self.att=a
        elif isinstance(X,Ball):
            self.b=X; self.Type=self.BALL; self.att=a
        else:
            self.c=X; self.Type=self.FREE; self.att=a
#■表示しようとシーンに、どのような物体があるかを表現するクラス
class Scene(Attribute):
    MAXOBJ      = 1024
    WALL        = 0x10
    BALL        = 0x20
    FREE        = 0x40
    BVOL        = 0xF0
    NOINTERSECT = 1E8
    Object = [0 for i in range(MAXOBJ)] #物体
    n   = 0                             #物体の個数
    bgc = Color(0,0,0)                  #背景色
    md  = False                         #表示の有無
    def __init__(self, C = Color(0,0,0), M=True):
        self.n=0; self.md=M; self.bgc=C
    def bgcolor(self):return self.bgc
    def regist(self, a, c, b=0):
        if isinstance(a, Wall) or isinstance(a, Ball):
            self.Object[self.n]=Object(a,c)
            self.n+=1
    def Attribute(self, i):
        return (self.Object[i]).att
    def hit(self,VR):
        t=self.NOINTERSECT; tt=0.0
        HP=Point(0,0,0); HN=Point(0,0,0)
        P=HP; N=HN; hi=-1
        for i in range(self.n):
            if self.Object[i].Type==self.WALL :
                RR=self.Object[i].w.hit(VR)
                tt=RR[0]
                if tt <self.NOINTERSECT: HP=RR[1]; HN=RR[2]
            elif(self.Object[i].Type==self.BALL):
                RR=self.Object[i].b.hit(VR)
                tt=RR[0]
                if tt <self.NOINTERSECT: HP=RR[1]; HN=RR[2]
            if tt>0.5 and tt<t : t=tt; hi=i; P=HP; N=HN
        return [hi, P, N, hi]
    def light(self, VR,L) : #照明光の強さ
        s=1.0; tt=0.0
        HP=Point(0,0,0);HN=Point(0,0,0)
        for i in range(self.n):
            if self.Object[i].Type==self.WALL :
                RR=self.Object[i].w.hit(VR)
                tt=RR[0]
                if tt <self.NOINTERSECT: HP=RR[1]; HN=RR[2]
            elif self.Object[i].Type==self.BALL:
                RR=self.Object[i].b.hit(VR);
                tt=RR[0]
                if tt <self.NOINTERSECT: HP=RR[1]; HN=RR[2]
            if tt>0.5 and tt<L-0.5:
                if self.Object[i].att.kn<=0 : s=0
                else: s *= (1-self.Object[i].att.kr)
            if(s<=0) :break
        return s
    
    def color(self, VR, L0, rn, am=1):
        if (rn==0) or (am<0.01): return [Color(0,0,0), L0, rn]
        Ps=Point(0,0,0);  N=Point(0,0,0)
        hi=-1; rnt=rn-1; rnr=rn-1   # hi: 視線が交わった物体の番号
        RR=self.hit(VR)
        if RR[0]<0 : return [self.bgcolor(),L0,rn]
        Ps=RR[1]; N=RR[2]; hi=RR[3]
        Cs=Color(0,0,0); Cr=Color(0,0,0); Ct=Color(0,0,0)
        A  = self.Attribute(hi)
        kr = A.kr
        L  = unit(Ps-L0)                    #照明光方向ベクトル
        LR = Ray(Ps,-L)                     #照明光線定義
        s  = self.light(LR, length(L0-Ps))  #交点が日陰
        Cs=shading(VR.d,L,N,A.c,s)          #交点自身の色
        if A.kn>0:                          #屈折光処理
            n=A.kn
            if N*VR.d>0 : n=1/n; N=-N       #物体から出る場合
            c = -VR.d*N ; g2 = n*n+c*c-1
            if g2>0:                        #全反射でない場合
                g  = float(math.sqrt(float(g2)))
                TR = Ray(Ps, (VR.d+N*(c-g))/n)
                cmg=c-g; cpg=c+g; nmg=n*n*c-g; npg=n*n*c+g
                kr = float((cmg*cmg/(cpg*cpg)+nmg*nmg/(npg*npg))/2)
                RX = self.color(TR,L0,rnt, am*(1-kr)*A.kt)
                Ct = RX[0]; L0=RX[1]; rnt=RX[2]
            else: kr=1                      #全反射
        if kr>0 :
            if N*VR.d>0 : N=-N
            RR = Ray(Ps,VR.d-N*2*(VR.d*N))  # 反射視線定義
            RX = self.color(RR,L0,rnr,am*kr)
            Cr = RX[0]; L0=RX[1]; rnr=RX[2] # 反射視点から見える物体色
        rn=rnr
        if rnr<=rnt : rn=rnt
        rn+=1
        return [self.mixColor(Cr,Ct,Cs,kr,(1-kr)*A.kt,1),L0, rn]
    #-----------------------------------------------------------------
    #■色の混合
    #
    #　【注意】
    #     　本来の色混合は，各色のエネルギーの対数の加算で求めます。
    #     ただし，この方法では，計算量が多くなりますので，
    #     本プログラムでは，筆者が1990年代に考えた簡易な方法を採用しています。
    #
    #       現実の色に近づけるためには本来の方法を使うべきでしょうが，
    #     この簡易な方法でも，近似的な，それなりの色になります。
    #
    #       自然色に近づけたい方は，この部分の改良をお願いします。
    #
    def mixColor(self, c1, c2, c3, a1, a2, a3):
        R1= float(c1.R); G1= float(c1.G); B1= float(c1.B)
        R2= float(c2.R); G2= float(c2.G); B2= float(c2.B)
        R3= float(c3.R); G3= float(c3.G); B3= float(c3.B)
        R= checkRange(int(R1*a1+R2*a2+R3*a3), 0, 255)
        G= checkRange(int(G1*a1+G2*a2+G3*a3), 0, 255)
        B= checkRange(int(B1*a1+B2*a2+B3*a3), 0, 255)
        return Color(R,G,B)
